﻿using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;
using System.Runtime.CompilerServices;

BenchmarkRunner.Run<InliningBenchmarks>();

public class A {
	public int X;

	public int GetX() {
		return X;
	}

	public void SetX(int value) {
		X = value;
	}

	public void SetNonNegativeX(int value) {
		if (value < 0) throw new ArgumentOutOfRangeException("Too small", nameof(value));
		if (value > 1_000_000) throw new ArgumentOutOfRangeException("Too big", nameof(value));
		if (value == 0xC00FEE) throw new ArgumentOutOfRangeException("Too strange", nameof(value));
		X = value;
	}
}

public class InliningBenchmarks {
	[Params(42)]
	public int Param;

	public static A a1 = new A();
	public static A a2 = new A();

	[Benchmark]
	public int A_Just1FieldRead() {
		return a1.X;
	}

	[Benchmark]
	public int A_Just1GetX() {
		return a1.GetX();
	}

	[Benchmark]
	public void A_Just1SetX() {
		a1.SetX(Param);
	}

	[Benchmark]
	public void A_Just1SetNonNegativeX() {
		a1.SetNonNegativeX(Param);
	}

	[Benchmark]
	public int A_Multi_Field() {
		a1.X = Param;
		a2.X = Param;

		while (a1.X > 0) {
			int temp = a2.X;
			a2.X = a1.X - 1;
			a1.X = temp;
		}

		return a1.X + a2.X;
	}


	[Benchmark]
	public int A_Multi_GetSetX() {
		a1.SetX(Param);
		a2.SetX(Param);

		while (a1.GetX() > 0) {
			int temp = a2.GetX();
			a2.SetX(a1.GetX() - 1);
			a1.SetX(temp);
		}

		return a1.GetX() + a2.GetX();
	}

	[Benchmark]
	public int A_Multi_GetSetNonNegativeX() {
		a1.SetNonNegativeX(Param);
		a2.SetNonNegativeX(Param);

		while (a1.GetX() > 0) {
			int temp = a2.GetX();
			a2.SetNonNegativeX(a1.GetX() - 1);
			a1.SetNonNegativeX(temp);
		}

		return a1.GetX() + a2.GetX();
	}

	public static B b1 = new B();
	public static B b2 = new B();

	[Benchmark]
	public int B_Just1GetX() {
		return b1.GetX();
	}

	[Benchmark]
	public void B_Just1SetX() {
		b1.SetX(Param);
	}

	[Benchmark]
	public void B_Just1SetNonNegativeX() {
		b1.SetNonNegativeX(Param);
	}

	[Benchmark]
	public int B_Multi_Field() {
		b1.X = Param;
		b2.X = Param;

		while (b1.X > 0) {
			int temp = b2.X;
			b2.X = b1.X - 1;
			b1.X = temp;
		}

		return b1.X + b2.X;
	}

	[Benchmark]
	public int B_Multi_GetSetX() {
		b1.SetX(Param);
		b2.SetX(Param);

		while (b1.GetX() > 0) {
			int temp = b2.GetX();
			b2.SetX(b1.GetX() - 1);
			b1.SetX(temp);
		}

		return b1.GetX() + b2.GetX();
	}

	[Benchmark]
	public int B_Multi_GetSetNonNegativeX() {
		b1.SetNonNegativeX(Param);
		b2.SetNonNegativeX(Param);

		while (b1.GetX() > 0) {
			int temp = b2.GetX();
			b2.SetNonNegativeX(b1.GetX() - 1);
			b1.SetNonNegativeX(temp);
		}

		return b1.GetX() + b2.GetX();
	}

	public static C c1 = new C();
	public static C c2 = new C();

	[Benchmark]
	public int C_Just1GetX() {
		return c1.GetX();
	}

	[Benchmark]
	public void C_Just1SetX() {
		c1.SetX(Param);
	}

	[Benchmark]
	public void C_Just1SetNonNegativeX() {
		c1.SetNonNegativeX(Param);
	}

	[Benchmark]
	public int C_Multi_Field() {
		c1.X = Param;
		c2.X = Param;

		while (c1.X > 0) {
			int temp = c2.X;
			c2.X = c1.X - 1;
			c1.X = temp;
		}

		return c1.X + c2.X;
	}

	[Benchmark]
	public int C_Multi_GetSetX() {
		c1.SetX(Param);
		c2.SetX(Param);

		while (c1.GetX() > 0) {
			int temp = c2.GetX();
			c2.SetX(c1.GetX() - 1);
			c1.SetX(temp);
		}

		return c1.GetX() + c2.GetX();
	}

	[Benchmark]
	public int C_Multi_GetSetNonNegativeX() {
		c1.SetNonNegativeX(Param);
		c2.SetNonNegativeX(Param);

		while (c1.GetX() > 0) {
			int temp = c2.GetX();
			c2.SetNonNegativeX(c1.GetX() - 1);
			c1.SetNonNegativeX(temp);
		}

		return c1.GetX() + c2.GetX();
	}
}

public class B {
	public int X;

	[MethodImpl(MethodImplOptions.AggressiveInlining)]
	public int GetX() {
		return X;
	}

	[MethodImpl(MethodImplOptions.AggressiveInlining)]
	public void SetX(int value) {
		X = value;
	}

	[MethodImpl(MethodImplOptions.AggressiveInlining)]
	public void SetNonNegativeX(int value) {
		if (value < 0) throw new ArgumentOutOfRangeException("Too small", nameof(value));
		if (value > 1_000_000) throw new ArgumentOutOfRangeException("Too big", nameof(value));
		if (value == 0xC00FEE) throw new ArgumentOutOfRangeException("Too strange", nameof(value));
		X = value;
	}
}

public class C {
	public int X;

	[MethodImpl(MethodImplOptions.NoInlining)]
	public int GetX() {
		return X;
	}

	[MethodImpl(MethodImplOptions.NoInlining)]
	public void SetX(int value) {
		X = value;
	}

	[MethodImpl(MethodImplOptions.NoInlining)]
	public void SetNonNegativeX(int value) {
		if (value < 0) throw new ArgumentOutOfRangeException("Too small", nameof(value));
		if (value > 1_000_000) throw new ArgumentOutOfRangeException("Too big", nameof(value));
		if (value == 0xC00FEE) throw new ArgumentOutOfRangeException("Too strange", nameof(value));
		X = value;
	}
}
